static char rcsid[] = "@(#)$Id: init.c,v 2.11 2015/04/22 14:53:02 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.11 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> (was hurtta+elm@posti.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/


#include "def_misc.h"
#include "s_elm.h"
#include "conf_writer_imp.h"
#include "save_opts.h"
#include "rc_imp.h"


DEBUG_VAR(Debug,__FILE__,"mime");

#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

S_(dump_conf_map_f dump_user_iso2022_sets)
static void dump_user_iso2022_sets P_((FILE *f,const char *actor,
				       char *version_buff));
static void dump_user_iso2022_sets(f,actor,version_buff) 
     FILE *f;
     const char *actor;
     char *version_buff;
{
    dump_iso2022_map(f,ml_user);
}

S_(dump_message_f user_iso2022_sets_written)
static void user_iso2022_sets_written P_((char *fname));
static void user_iso2022_sets_written(fname)
     char *fname;
{
    lib_error(CATGETS(elm_msg_cat, ElmSet, 
		      ElmISO2022SavedIn,
		      "ISO 2022 mappings saved in file %s."), 
	      fname);		    
}

S_(dump_conf_map_f dump_user_mime_charsets)
static void dump_user_mime_charsets P_((FILE *f,const char *actor,
					char *version_buff));
static void dump_user_mime_charsets(f,actor,version_buff) 
     FILE *f;
     const char *actor;
     char *version_buff;
{
    dump_charset_map(f,user_charset_map,NULL,actor,version_buff,0);
}

S_(dump_message_f user_mime_charsets_written)
static void user_mime_charsets_written P_((char *fname));
static void user_mime_charsets_written(fname)
     char *fname;
{
    lib_error(CATGETS(elm_msg_cat, ElmSet, 
		      ElmCharsetsSavedIn,
		      "Charsets saved in file %s."), 
	      fname);
}

S_(dump_conf_map_f dump_user_terminal_info)
static void dump_user_terminal_info P_((FILE *f,const char *actor,
					char *version_buff));
static void dump_user_terminal_info(f,actor,version_buff) 
     FILE *f;
     const char *actor;
     char *version_buff;
{
    dump_terminal_map(f,user_terminal_map,NULL,actor,version_buff);
}

S_(dump_message_f user_terminal_info_written)
static void user_terminal_info_written P_((char *fname));
static void user_terminal_info_written(fname)
     char *fname;
{
    lib_error(CATGETS(elm_msg_cat, ElmSet, 
		      ElmTerminalInfoSavedIn,
		      "Additional terminal info saved in file %s."), 
	      fname);
}

S_(dump_conf_map_f dump_user_mime_types)
static void dump_user_mime_types P_((FILE *f,const char *actor,
				     char *version_buff));
static void dump_user_mime_types(f,actor,version_buff) 
     FILE *f;
     const char *actor;
     char *version_buff;
{
    dump_mime_types_map(f,user_mimetypes_map,NULL,actor,version_buff);
}

S_(dump_message_f user_mime_types_written)
static void user_mime_types_written P_((char *fname));
static void user_mime_types_written(fname)
     char *fname;
{
    lib_error(CATGETS(elm_msg_cat, ElmSet, 
		      ElmContentTypeInfoSavedIn,
		      "Content-type mapping info saved in file %s."), 
	      fname);
}

S_(dump_conf_map_f dump_user_mlist_file)
static void dump_user_mlist_file P_((FILE *f,const char *actor,
				     char *version_buff));
static void dump_user_mlist_file(f,actor,version_buff) 
     FILE *f;
     const char *actor;
     char *version_buff;
{
    dump_mlist_conf(f,user_mlist_conf,NULL,actor,version_buff,
		    user_mlist_cs,user_mlist_pl);
}

S_(dump_message_f user_mlist_file_written)
static void user_mlist_file_written P_((char *fname));
static void user_mlist_file_written(fname)
     char *fname;
{
    lib_error(CATGETS(elm_msg_cat, ElmSet, 
		      ElmMailingListInfoSavedIn,
		      "Mailing lists info saved in file %s."), 
	      fname);
}

struct terminal_map_item * system_terminal_map;
struct terminal_map_item * user_terminal_map;
struct mime_types_item   * system_mimetypes_map;
struct mime_types_item   * user_mimetypes_map;

struct mlist_conf        * system_mlist_conf;
charset_t                  system_mlist_cs;
struct editor_propline   * system_mlist_pl;
struct mlist_conf        * user_mlist_conf;
charset_t                  user_mlist_cs;
struct editor_propline   * user_mlist_pl;

static char * EDITOR_PROPLINE [] = 
    { "auto", "emacs", "ignore", NULL };
ENUMERATE editor_ml_propline = {
    editor_propline_auto,
    NUM_editor_propline,
    &(EDITOR_PROPLINE[0])
};

static ZZZ_SAVE_TYPE misc_info_data[] = {
    {"editor-ml-propline",   ZZZ_DT_ENUM(&editor_ml_propline), 0 },
};
static int COUNT=(sizeof(misc_info_data)/sizeof(ZZZ_SAVE_TYPE));
static save_info_recs *misc_info = (save_info_recs *) misc_info_data;

#if ANSI_C
static init_default_hook init_misc_defaults;
#endif
static void init_misc_defaults P_((int *errors));
static void init_misc_defaults(errors)
     int *errors;
{
    system_terminal_map  = load_terminal_map(system_terminal_info,errors);
    system_mimetypes_map = load_mime_types_map(system_mime_types,errors);
    system_mlist_conf    = load_mlist_conf(system_mlist_file,errors,
					   &system_mlist_cs,
					   &system_mlist_pl);

    register_conf_write(USER_ISO2022_SETS,user_iso2022_sets,
			dump_user_iso2022_sets,
			user_iso2022_sets_written);
    register_conf_write(USER_MIME_CHARSETS,user_mime_charsets,
			dump_user_mime_charsets,
			user_mime_charsets_written);
    register_conf_write(USER_TERMINAL_INFO,user_terminal_info,
			dump_user_terminal_info,
			user_terminal_info_written);   
    register_conf_write(USER_MIME_TYPES,user_mime_types,
			dump_user_mime_types,
			user_mime_types_written);
    register_conf_write(USER_MLIST_FILE,user_mlist_file,
			dump_user_mlist_file,
			user_mlist_file_written);
}

#if ANSI_C
static init_default_hook misc_read_rc_file;
#endif
static void misc_read_rc_file P_((int *errors));
static void misc_read_rc_file(errors)
     int *errors;
{
    user_terminal_map  = load_terminal_map(user_terminal_info,errors);
    user_mimetypes_map = load_mime_types_map(user_mime_types,errors);
    user_mlist_conf    = load_mlist_conf(user_mlist_file,errors,
					 &user_mlist_cs,&user_mlist_pl);
}

void init_misclib()
{
    register_delayed(misc_info,COUNT);
    add_init_default_hook(&init_misc_defaults);
    add_read_rc_file_hook(&misc_read_rc_file);
}



/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
