static char rcsid[] = "@(#)$Id: elmconfwriter.c,v 2.3 2015/04/22 14:53:03 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.3 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *       or Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "def_utils.h"
#include "s_me.h"
#include "s_elm.h"
#include "s_elmrc.h"

#include "rc_imp.h"
#include "save_opts.h"

#include "misclib.h"
#include "addrlib.h"
#include "aliaslib.h"

#include "reghelper.h"

DEBUG_VAR(Debug,__FILE__,"util");

extern char *optarg;		
extern int   optind;		

int    register_fd     = -1; 
const char * program_name    = "elmconfwriter"; 
char * register_module = NULL;

static char version_buff[NLEN];

static const char OPTION_LETTERS[] = "d:c:l";

int main P_((int argc, char *argv[]));
int main(argc, argv)
     int argc;
     char *argv[];
{
    int err = 0;
    int c;
    char * elmrc_changes = NULL;
    struct elmrc_recorder * recorder = NULL;
    FILE * elmrc_changes_F = NULL;
    int list_files = 0;

    enum {
	conf_elmrc,
	conf_charset_map,
	conf_mime_types_map,
	conf_terminal_map,
	conf_iso2022_map,
	conf_mail_services,
	conf_aliases_map,
	conf_mlist_map,
	conf_LAST
    } config_file;
    
#if DEBUG
    init_debugfile("ELMCONFWRITER");
#endif
    
    while ((c = ELM_GETOPT(argc, argv, OPTION_LETTERS)) != EOF) {
	switch(c) {
	case 'd' : 	
#if DEBUG
	    set_debugging(optarg);	  
#endif
	    /* Error is printed later */
	    break;
	}
    }
    optind = 1;     /* Reset scanning */

    locale_init();

    REGHELPER_INIT(argv[0]);

while ((c =  ELM_GETOPT(argc, argv, OPTION_LETTERS)) != EOF) {
	
	switch(c) {
	    
	case 'd':
#if DEBUG
	    set_debugging(optarg);	 
#else
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmArgsIngoringDebug,
			      "Warning: system created without debugging enabled - request ignored\n"));
#endif
	    break;
	    

	case 'c':
	    elmrc_changes = optarg;
	    if (0 != access(elmrc_changes,WRITE_ACCESS)) {
		int errcode = errno;
		
		if (errcode != ENOENT) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
				      "File %.50s is not writeable: %s"),
			      elmrc_changes, strerror(errcode));
		    err=1;
		    goto fail;
		}
	    }


	    elmrc_changes_F = open_or_create(elmrc_changes);
	    if (!elmrc_changes_F) {
		int errcode = errno;
		
		if (errcode != ENOENT) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
				      "File %.50s is not writeable: %s"),
			      elmrc_changes, strerror(errcode));
		    err=1;
		    goto fail;
		}
	    }

	    break;

        case 'l':
	    list_files = 1;
	    break;

	case '?':
	    err = 1;
	    goto fail;
	}

    }

    user_init();

    init_misclib();

    if (elmrc_changes_F) {
	recorder = enable_recorder(SYSTEM_RC);
	
	seed_history(elmrc_changes_F,elmrc_changes,recorder);
    }

    init_aliaslib();

    init_defaults(READ_FLAG_GLOBAL_ONLY);

    elm_sfprintf(version_buff, sizeof version_buff,
		 FRM("%s PL%s"), VERSION, PATCHLEVEL);
    
#ifdef DEBUG
    { 
	int d = panic_dprint("\n\
======================================================\n\
Debug output of the ELMCONFWRITER program (version %s).\n",
			     version_buff);

	if (d >= 50) {
#if 0	
	    panic_dprint("WARNING: Edit manually out sensitive information from that file!\n");
	    
	    lower_prompt("WARNING: Debug file may include passwords -- edit it!");
	    sleep(5+sleepmsg);	    
#endif
	}
    }
#endif
    
    post_init_check(0);


    for (config_file = conf_elmrc; config_file < conf_LAST; config_file++) {
	char * targetfile = NULL;

	switch (config_file) {
	case conf_elmrc: targetfile = system_rc_file;
	    break;
	case conf_charset_map: targetfile = system_mime_charsets;
	    break;
	case conf_mime_types_map: targetfile = system_mime_types;
	    break;
	case conf_terminal_map: targetfile = system_terminal_info;
	    break;
	case conf_iso2022_map: targetfile = system_iso2022_sets;
	    break;
	case conf_mail_services: targetfile = system_mail_services;
	    break;
	case conf_aliases_map: targetfile = system_aliases_file;
	    break;
	case conf_mlist_map: targetfile = system_mlist_file;
	    break;

	case conf_LAST:
	    break;
	}

	if (targetfile) {
	    int r = 0;
	    FILE * f = NULL;
	    char * tmp = NULL;
	    struct stat target_stat;
	    int mode = 0644;

	    if (0 != access(targetfile,WRITE_ACCESS)) {
		int errcode = errno;
		
		if (ENOENT == errcode) {
		    
		    if (list_files)
			printf ("%s (does not exists)\n",
				targetfile);
		    
		    continue;
		}
		
		if (list_files)
		    printf ("%s (error: %s)\n",
			    targetfile,strerror(errcode));
		else
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
				      "File %.50s is not writeable: %s"),
			      targetfile, strerror(errcode));
		err=1;
		continue;
	    }
	    
	    if (list_files) {
		printf ("%s\n", targetfile);
		continue;
	    }

	    if (0 != stat(targetfile,&target_stat)) {
		int errcode = errno;

		lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNoStat,
				  "File %.30s: %.40s"),
			  targetfile, strerror(errcode));
		err=1;
		continue;			  
	    }
	    mode = target_stat.st_mode & 07777;

	    if (0 != (mode & 00002)) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeFileGlobWritable,
				  "Config file %s is globally writable, access mode %05o"), 
			  targetfile,mode);
		err=1;	  
	    }

	    if (0 == (mode & 00004)) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeFileGlobUnreadable,
				  "Config file %s is not globally readable, access mode %05o"),
			  targetfile,mode);
		err=1;
	    } else
		mode |= 0444;

	    switch (config_file) {
	    case conf_charset_map:
	    case conf_mime_types_map:
	    case conf_terminal_map: 
	    case conf_iso2022_map:
	    case conf_mail_services:
	    case conf_aliases_map: 
	    case conf_mlist_map:

		tmp = elm_message(FRM("%s.N"),targetfile);
		f = open_or_create(tmp);
		
		if (!f) {
		    int errcode = errno;
		    
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
				      "File %.50s is not writeable: %s"),
			      tmp, strerror(errcode));
		    
		    err=1;
		    free(tmp);
		    continue;
		}
		
		break;
		
	    case conf_elmrc:
	    case conf_LAST:
		break;
	    }


	    switch (config_file) {
	    case conf_elmrc:
		r = write_rc(targetfile,NULL,1,
			     "ELMCONFWRITER", NULL,
			     version_buff);
		break;

	    case conf_charset_map:
		dump_charset_map(f,system_charset_map,NULL,
				 "ELMCONFWRITER",version_buff,0);
		r = 1;
		break;
		
	    case conf_mime_types_map:
		dump_mime_types_map(f,system_mimetypes_map,NULL,
				    "ELMCONFWRITER",version_buff);
		r = 1;
		break;

	    case conf_terminal_map: 
		dump_terminal_map(f,system_terminal_map,NULL,
				  "ELMCONFWRITER",version_buff);
		r = 1;
		break;

	    case conf_iso2022_map:
		dump_iso2022_map(f,ml_global);
		r = 1;
		break;

	    case conf_mail_services:
		r = dump_service_entries(f,system_mail_services_conf,NULL,
				     "ELMCONFWRITER",version_buff,
				     system_mail_services_cs,
				     system_mail_services_pl);
		break;

	    case conf_aliases_map: 
		dump_aliases_map(f,system_aliases_map,NULL,
				 "ELMCONFWRITER",version_buff,
				 system_aliases_cs,
				 system_aliases_pl);
		r = 1;
		break;

	    case conf_mlist_map:
		dump_mlist_conf(f,system_mlist_conf,NULL,
				"ELMCONFWRITER",version_buff,
				system_mlist_cs,system_mlist_pl);
		r = 1;
		break;

	    case conf_LAST:
		break;
	    }

	    switch (config_file) {
	    case conf_charset_map:
	    case conf_mime_types_map:
	    case conf_terminal_map:
	    case conf_iso2022_map: 
	    case conf_mail_services:
	    case conf_aliases_map: 
	    case conf_mlist_map:

		if (EOF == fclose(f)) {
		    int errcode = errno;
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
				      "File %.50s is not writeable: %s"),
			      tmp, strerror(errcode));
		    
		    err=1;
		    free(tmp);
		    continue;
		}

		if (0 != rename(tmp,targetfile)) {
		    int errcode = errno;
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotRenamed,
				      "Failed to rename temporary file to %.50s: %.30s"),
			      targetfile, strerror(errcode));
	    
		    err=1;
		    free(tmp);
		    continue;
		}

		free(tmp);

		break;

	    case conf_elmrc:
	    case conf_LAST:
		break;
	    }
	    
	    if (!r) {
		err=1;
		continue;
	    }
	
	    if (0 != chmod(targetfile,mode)) {
		int errcode = errno;
		lib_error(CATGETS(elm_msg_cat, MeSet, MeFileModeNotChanged,
				  "Failed to change file %s mode (new mode %05o):  %s"),
			  targetfile,mode, strerror(errcode));
		err=1;
	    }
	
	    log_config(targetfile);
	}
    }


    if (elmrc_changes_F) {
	rewind(elmrc_changes_F);

#ifdef FTRUNCATE
	/* Not really necessary */
	if (-1 == ftruncate(fileno(elmrc_changes_F),0)) {
	    DPRINT(Debug,27,(&Debug,"ftruncate of %s failed\n",elmrc_changes));
	}
#endif

	print_history_changed(elmrc_changes_F,recorder);

    }

 fail:
    if (elmrc_changes_F) 
	fclose(elmrc_changes_F);

    if (err)
	lib_error(CATGETS(elm_msg_cat, MeSet, MeProgFailed,
			  "%s failed; exit code=%d"),
		  argv[0],err);

    return err;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
